/////////////////////
/// author Yi Gu
///
/// Clique simplification Algorithm:
///
/// Given a graph, we would like to detect
/// the clique structures

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <set>
#include <queue>
#include <map>
#include <vector>

using namespace std;

//------------------------------------------------------------
/// This function looks for the node with the largest degree
/// subG: graph
/// adjMat: adjacent matrix
/// num_nodes: number of nodes
int findNode_LargestDegree(set<int> subG, bool* adjMat, unsigned int num_nodes);

//------------------------------------------------------------
/// This function looks for the leading nodes of the potential cliques
/// subG: graph
/// candicatesInCurrLevel: candicates
/// adjMat: adjacent matrix
/// num_nodes: number of nodes
int candidateNodeDetection(set<int> subG, queue<int> *candicatesInCurrLevel,
                            bool* adjMat, unsigned int num_nodes);

//------------------------------------------------------------
/// This function expands the nodes. The expanding history forms
/// a clique
/// subG: graph
/// cand: candidates
/// his: expanding history
/// isFirstLevel: whether this expanding is the first level
/// firstLevelNodes: the nodes in the first level
/// cliques: the structure that stores the cliques
/// expandNum: how many levels have been expanded
/// adjMat: adjacent matrix
/// num_nodes: number of nodes
void expand(set<int> subG, set<int> cand, vector<int> *his,
                      bool isFirstLevel, set<int> firstLevelNodes,
                      vector<vector<int> > *cliques, int* expandNum,
                      bool* adjMat, unsigned int num_nodes);

//------------------------------------------------------------
/// This function finds the cliques.
/// adjMat: adjacent matrix
/// num_nodes: number of nodes
/// cliques: the structure that stores the cliques
void cliqueSimp(bool* adjMat, unsigned int num_nodes, vector<vector<int> > *cliques);

//------------------------------------------------------------
/// This function reads in the adjacency matrix.
/// fn: file name
/// adjMat: adjacent matrix
/// num_nodes2: num_nodes * num_nodes
void readGraph(char* fn, bool* adjMat, unsigned int num_nodes2);

//------------------------------------------------------------
/// This function stores the cliques that we find.
/// fn: file name
/// cliques: the structure that stores the cliques
void storeConnectors(char* fn, vector<vector<int> > cliques);
