#include "connector.h"

void storeEdge4EachNode(unsigned int num_nodes,
                                  vector<vector<unsigned int> > *incomingEdges,
                                  bool* adjMat){
    int i, j;

    for(j = 0; j < num_nodes; ++j){
        (*incomingEdges).push_back(vector<unsigned int>());

        for(i = 0; i < num_nodes; ++i){
            if(i == j){
                // ignore
            }else if(adjMat[i * num_nodes + j]){
                (*incomingEdges)[j].push_back(i);
            }
        }
    }
}

void connectorSimp(multimap<int, int> *connectors, set<int> *keys,
                   unsigned int num_nodes, vector<vector<unsigned int> > incomingEdges){
    // the key is used to store the pair of starting and ending values
    // the value is the id of the middle node
    unsigned int i, key, left, right;

    for(i = 0; i < num_nodes; ++i){
        //	the middle node should not be the fan node, nor the fan member

        if(incomingEdges[i].size() == 2){
            // May be the middle node in a connector

            left = incomingEdges[i][0];

            right = incomingEdges[i][1];

            if(left < right){
                key = left * num_nodes + right;
            }else{
                key = right * num_nodes + left;
            }

            (*keys).insert(key);
            (*connectors).insert(pair<unsigned int, unsigned int>(key, i));
        }
    }
}

void readGraph(char* fn, bool* adjMat, unsigned int num_nodes2){
    FILE* pFile = fopen(fn, "rb");

    if(pFile != NULL){
        fread(adjMat, sizeof(bool), num_nodes2, pFile);

        fclose(pFile);
    }else{
        printf("Cannot open file %s\n", fn);
    }
}

void storeConnectors(char* fn, multimap<int, int> connectors, set<int> keys, unsigned int num_nodes){
    FILE* pFile = fopen(fn, "wt");

    if(pFile != NULL){
        set<int>::iterator it_keys;

        pair<multimap<int, int>::iterator, multimap<int, int>::iterator> sameRange;
        int id, left, right;

        for(it_keys = keys.begin(); it_keys != keys.end(); ++it_keys){
            unsigned int key = (*it_keys);

            left = key / num_nodes;
            right = key % num_nodes;

            if(connectors.count(key) >= 2){
                sameRange = connectors.equal_range(key);

                fprintf(pFile, "%d %d: ", left, right);

                for (multimap<int, int>::iterator it2 = sameRange.first; it2 != sameRange.second; ++it2){
                    id = (*it2).second;

                    fprintf(pFile, "%d ", id);
                }

                fprintf(pFile, "\n");
            }
        }

        fclose(pFile);
    }else{
        printf("Cannot open file %s\n", fn);
    }
}

int main(int argc, char* argv[]){
    char ifn[1024];
    char ofn[1024];

    unsigned int num_nodes, num_nodes2;

    if (argc != 4) {
        // Tell the user how to run the program
        printf("Input Error. Please input: the input and output data path, number of nodes, number of iterations and C2.");

        return 1;
    }else{
        for (unsigned int i = 1; i < 4; ++i) {
            // Remember argv[0] is the path to the program, we want from argv[1] onwards
           if(i == 1){
               strcpy(ifn, argv[i]);
           }else if(i == 2){
               strcpy(ofn, argv[i]);
           }else if(i == 3){
               num_nodes = atoi(argv[i]);
           }
       }
    }

    num_nodes2 = num_nodes * num_nodes;

    bool* adjMat = new bool[num_nodes2];
    memset(adjMat, false, sizeof(bool) * num_nodes2);

    readGraph(ifn, adjMat, num_nodes2);

    vector<vector<unsigned int> > incomingEdges;

    storeEdge4EachNode(num_nodes, &incomingEdges, adjMat);

    multimap<int, int> connectors;
    set<int> keys;

    connectorSimp(&connectors, &keys, num_nodes, incomingEdges);

    storeConnectors(ofn, connectors, keys, num_nodes);

    delete[] adjMat;

    connectors.clear();
    keys.clear();

    return 1;
}
