/////////////////////
/// author Yi Gu
///
/// Connector simplification Algorithm:
///
/// Given a graph, we would like to detect
/// the Connector structures

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <set>
#include <queue>
#include <map>
#include <vector>

using namespace std;

//------------------------------------------------------------
/// This function converts the adjacent matrix to a 2D vector
/// which make it easy to extract the nodes linked to a given
/// num_nodes: the number of nodes in a graph
/// incomingEdges: 2D vector that stores the information of edges
/// adjMat: adjacent matrix
void storeEdge4EachNode(unsigned int num_nodes,
                                  vector<vector<unsigned int> > *incomingEdges,
                                  bool* adjMat);

//------------------------------------------------------------
/// This function helps us identify connector structures
/// connectors: It stores connectors fan structures
/// keys: It stores the two ending nodes of all the connectors structures.
/// num_nodes: the number of nodes in a graph
/// incomingEdges: 2D vector that stores the information of edges
void connectorSimp(multimap<int, int> *connectors, set<int> *keys,
                   unsigned int num_nodes, vector<vector<unsigned int> > incomingEdges);

//------------------------------------------------------------
/// This function read an input graph in the form of adjacent matrix
/// fn: file name
/// adjMat: adjacent matrix
/// num_nodes2: num_nodes * num_nodes
void readGraph(char* fn, bool* adjMat, unsigned int num_nodes2);

//------------------------------------------------------------
/// This function stores the information of connectors
/// fn: file name
/// connectors: It stores the connectors structures
/// keys: It stores the two ending nodes of all the connectors structures.
void storeConnectors(char* fn, multimap<int, int> connectors, set<int> keys, unsigned int num_nodes);
