#include "simRank.h"

void storeEdge4EachNode(int num_nodes,
                                  vector<vector<pair<int, float> > > *incomingEdges,
                                  float* tranMat){
    int i, j;

    for(j = 0; j < num_nodes; ++j){
        (*incomingEdges).push_back(vector<pair<int, float> >());

        for(i = 0; i < num_nodes; ++i){
            if(i == j){
                // ignore
            }else if(tranMat[i * num_nodes + j] > ZERO){
                pair<int, float> p;

                p.first = i;
                p.second = tranMat[i * num_nodes + j];

                (*incomingEdges)[j].push_back(p);
            }
        }
    }
}

void computeSimRank(float* simRank, float* tranMat,
                              vector<vector<pair<int, float> > > incomingEdges,
                              int num_nodes, int
                              num_iters, float C2){
    memset(simRank, -1, sizeof(float) * num_nodes * num_nodes);

    int i, j, a, b, t;

    float sum, value, weight, totalWeight;
    int count;

    for(t = 0; t < num_iters; ++t){
        // for each iteration, the same node is 1
        value = 0;

        for(i = 0; i < num_nodes; ++i){
            for(j = 0; j < num_nodes; ++j){
                if(i == j){
                    simRank[i * num_nodes + j] = 1.0f;
                }else{
                    // i's neighbors and j's neighbors
                    sum = 0.0f;
                    totalWeight = 0.0f;
                    count = 0;

                    for(a = 0; a < incomingEdges[i].size(); ++a){
                        pair<int, float> p1 = incomingEdges[i][a];

                        pair<int, float> p2;

                        for(b = 0; b < incomingEdges[j].size(); ++b){
                            p2 = incomingEdges[j][b];

                            value = simRank[p1.first * num_nodes + p2.first];

                            weight = p1.second * p2.second;

                            if(value >= 0){
                                sum += value * weight;

                                count++;

                                totalWeight += weight;
                            }
                        }
                    }

                    if(count > 0){

                        if(totalWeight > ZERO){
                            simRank[i * num_nodes + j] = C2 * sum / totalWeight;
                        }else{
                            simRank[i * num_nodes + j] = 0.0f;
                        }
                    }
                }
            }
        }


    }

    float max;
    int id;

    for(i = 0; i < num_nodes; ++i){
        max = -1.0;
        id = -1;

        for(j = 0; j < num_nodes; ++j){
            if(i != j && simRank[i * num_nodes + j] > ZERO){
                if(simRank[i * num_nodes + j] > max){
                    max = simRank[i * num_nodes + j];
                    id = j;
                }
            }
        }
    }
}

void calSimRank(float* tranMat, float* simRank, int num_nodes, int num_iters, float C2){
    vector<vector<pair<int, float> > > incomingEdges;

    storeEdge4EachNode(num_nodes, &incomingEdges, tranMat);

    computeSimRank(simRank, tranMat, incomingEdges, num_nodes, num_iters, C2);

    incomingEdges.clear();
}

void readFile(char* fn, float* tranMat, unsigned int num_nodes2){
    FILE* pFile = fopen(fn, "rb");

    if(pFile != NULL){
        fread(tranMat, sizeof(float), num_nodes2, pFile);

        fclose(pFile);
    }else{
        printf("Cannot read file %s \n", fn);
    }
}

void saveFile(char* fn, float* simMat, unsigned int num_nodes2){
    FILE* pFile = fopen(fn, "wb");

    if(pFile != NULL){
        fwrite(simMat, sizeof(float), num_nodes2, pFile);

        fclose(pFile);
    }else{
        printf("Cannot read file %s \n", fn);
    }
}

int main(int argc, char* argv[]){
    char ifn[1024], ofn[1024];
    unsigned int num_nodes, num_nodes2, num_iters;

    float C2;

    if (argc != 6) {
        // Tell the user how to run the program
        printf("Input Error. Please input: the input and output data path, number of nodes, number of iterations and C2.");

        return 1;
    }else{
        for (unsigned int i = 1; i < 6; ++i) {
            // Remember argv[0] is the path to the program, we want from argv[1] onwards
           if(i == 1){
               strcpy(ifn, argv[i]);
           }else if(i == 2){
               strcpy(ofn, argv[i]);
           }else if(i == 3){
               num_nodes = atoi(argv[i]);
           }else if(i == 4){
               num_iters = atoi(argv[i]);
           }else if(i == 5){
               C2 = atof(argv[i]);
           }
       }
    }

    float* tranMat = new float[num_nodes2];
    float* simRank = new float[num_nodes2];

    readFile(ifn, tranMat, num_nodes2);

    calSimRank(tranMat, simRank, num_nodes, num_iters, C2);

    saveFile(ofn, simRank, num_nodes2);

    delete[] tranMat;
    delete[] simRank;

    return 1;
}

