/////////////////////
/// author Yi Gu
///
/// SimRank Algorithm:
///
/// Two nodes in the graph are similar if their neighbors
/// are similar.

#include <vector>
#include <set>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

using namespace std;

const float ZERO = 0.000001;

//------------------------------------------------------------
/// This function converts the linkage tranMat to a format that
/// all the nodes can easily track their neighbors and weights.
/// num_nodes: the number of nodes
/// incomingEdges: the edges of all the nodes
/// tranMat: input linkage matrix
void storeEdge4EachNode(int num_nodes,
                                  vector<vector<pair<int, float> > > *incomingEdges,
                                  float* tranMat);

//------------------------------------------------------------
/// This function calculates the SimRank for all the nodes
/// simRank: simRank similarity matrix
/// tranMat: input linkage matrix
/// incomingEdges: the edges of all the nodes
/// num_nodes: the number of nodes
/// num_iters: the number of iterations
/// C2: a parameter
void computeSimRank(float* simRank, float* tranMat,
                              vector<vector<pair<int, float> > > incomingEdges,
                              int num_nodes, int
                              num_iters, float C2);

//------------------------------------------------------------
/// This function calls storeEdge4EachNode and computeSimRank
/// tranMat: input linkage matrix
/// simRank: simRank similarity matrix
/// num_nodes: the number of nodes
/// num_iters: the number of iterations
/// C2: a parameter
void calSimRank(float* tranMat, float* simRank, int num_nodes, int num_iters, float C2);

//------------------------------------------------------------
/// This function loads the linkage of the nodes.
/// fn: the input file path and name
/// tranMat: the linkage matrix
/// num_nodes2: num_nodes * num_nodes
void readFile(char* fn, float* tranMat, unsigned int num_nodes2);

//------------------------------------------------------------
/// This function saves the simRank.
/// fn: the output file path and name
/// simMat: SimRank matrix
/// num_nodes2: num_nodes * num_nodes
void saveFile(char* fn, float* simMat, unsigned int num_nodes2);
