#include "stateClustering_MPI.h"

using namespace std;

/// \brief loadData
/// \param bins
/// \param t
/// Load the histogram of the corresponding time step
void loadData(float* bins, int t){
    char fn[256];

#ifdef EARTHQUAKE
    sprintf(fn, "%s/%04d.dat", HISTOGRAMPATH, t + 1);
#elif defined EARTHQUAKE_16_16_16
    sprintf(fn, "%s/%04d.dat", HISTOGRAMPATH, t + 1);
#elif defined EARTHQUAKE_32_32_16
    sprintf(fn, "%s/%04d.dat", HISTOGRAMPATH, t + 1);
#else
    sprintf(fn, "%s/%04d.dat", HISTOGRAMPATH, t);
#endif

    FILE* pFile;

    pFile = fopen(fn, "rb");
  if(pFile == NULL){
        printf("can not open file %s \n", fn);

        exit(1);
    }else{
        fread(bins, sizeof(float), BINSIZE * BLOCKNUMXYZ, pFile);

        fclose(pFile);
    }
}

int getGlobalId(int t, int index){
    return (t * BLOCKNUMXYZ + index);
}

float JSD(float his1[BINSIZE], float his2[BINSIZE]){
    int i;

    for(i = 0; i < BINSIZE; ++i){
        avgHis[i] = 0.5f * (his1[i] + his2[i]);
    }

    float value = 0.0f;

    for(i = 0; i < BINSIZE; ++i){

        if(fabs(his1[i]) < ZERO && fabs(his2[i]) < ZERO){

        }else if(fabs(his1[i] < ZERO)){
            value += his2[i] * log(his2[i] / avgHis[i]);
        }else if(fabs(his2[i]) < ZERO){
            value += his1[i] * log(his1[i] / avgHis[i]);
        }else{
            value += his1[i] * log(his1[i] / avgHis[i]);
            value += his2[i] * log(his2[i] / avgHis[i]);
        }
    }

    value *= halfoneoverlog2;

    return value;
}

bool calculateHistogram(float his1[BINSIZE], float his2[BINSIZE], float threshold){
    float v = JSD(his1, his2);

    if(v <= threshold){
        return true;
    }

    return false;
}

void compare(float* targetHis, deque<blockInfo> *blockQueue, int targetTime, int targetIndex){
    deque<blockInfo>::iterator it;

    for(it = (*blockQueue).begin(); it != (*blockQueue).end(); ++it){
        (*it).similar = calculateHistogram(targetHis, (*it).his, THRESHOLD);
    }
}

void compare(float* targetHis, deque<blockInfo> *blockQueue){
    deque<blockInfo>::iterator it;

    for(it = (*blockQueue).begin(); it != (*blockQueue).end(); ++it){
        (*it).similar = calculateHistogram(targetHis, (*it).his, THRESHOLD);
    }
}

void getTargetTimeAndLastTimeInWindow(int targetGid, int* targetTime, int* targetIndex, int* lastTimeInWindow){
    (*targetTime) = targetGid / BLOCKNUMXYZ;
    (*targetIndex) = targetGid - (*targetTime) * BLOCKNUMXYZ;

    (*lastTimeInWindow) = (*targetTime) + TIMEWINDOW - 1;

    if((*lastTimeInWindow) >= BLOCKNUMT){
        (*lastTimeInWindow) = BLOCKNUMT - 1;
    }
}

void removeOldBlocks(deque<blockInfo> *blockQueue, int targetGid){
    blockInfo tempBlockInfo;

    while(!(*blockQueue).empty()){
        tempBlockInfo = (*blockQueue).front();

        if(getGlobalId(tempBlockInfo.t, tempBlockInfo.lindex) < targetGid){
            (*blockQueue).pop_front();
        }else{
            break;
        }
    }
}

void addNewBlocks(deque<blockInfo> *blockQueue, int lastTimeInWindow, int* blockIDs, int num_blocks, int* lastLoadedTime){    
    temp_start_t = MPI_Wtime();

    float temp_t = 0;

    int i, j;

    if((*lastLoadedTime) < lastTimeInWindow){
        // load data
        float* bins = new float[BLOCKNUMXYZB];

        for(i = (*lastLoadedTime) + 1; i <= lastTimeInWindow; ++i){

            start_t = MPI_Wtime();

            loadData(bins, i);

            end_t = MPI_Wtime();

            temp_t += end_t - start_t;

            for(j = 0; j < num_blocks; ++j){
                blockInfo newInfo;
                newInfo.t = i;
                newInfo.lindex = blockIDs[j];

                memcpy(newInfo.his, &(bins[BINSIZE * newInfo.lindex]), sizeof(float) * BINSIZE);
                (*blockQueue).push_back(newInfo);
            }
        }

        delete[] bins;

        (*lastLoadedTime) = lastTimeInWindow;
    }

    temp_end_t = MPI_Wtime();

    loading_t += temp_t;

    computing_t += temp_end_t - temp_start_t - temp_t;
}

bool haveTarget(deque<blockInfo> blockQueue, int targetTime, int targetIndex, float* targetHis){
    blockInfo head = blockQueue.front();

    if(head.t == targetTime){
        if(head.lindex == targetIndex){
            memcpy(targetHis, head.his, sizeof(float) * BINSIZE);

            return true;
        }
    }

    return false;
}

void broadCastHis(int rank, int num_proc, bool haveTarget, float* targetHis){
    MPI_Status status;

    int i;

    if(haveTarget){
        MPI_Request* request = new MPI_Request[num_proc];

        for(i = 0; i < num_proc; ++i){
            if(i != rank){
                MPI_Isend(targetHis, BINSIZE, MPI_FLOAT, i, i, MPI_COMM_WORLD, &(request[i]));
            }
        }

        for(i = 0; i < num_proc; ++i){
            if(i != rank){
                MPI_Wait(&(request[i]), &status);
            }
        }

        delete[] request;

    }else{
        MPI_Recv(targetHis, BINSIZE, MPI_FLOAT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
    }
}

///////// Section breaker *********
/// int MPI_Send(void *buf, int count, MPI_Datatype datatype, int dest, int tag,
/// MPI_Comm comm)
/// int MPI_Recv(void *buf, int count, MPI_Datatype datatype, int source, int tag,
/// MPI_Comm comm, MPI_Status *status)
///
/// \brief report2Master
/// \param rank
/// \param num_procs
/// \param blockQueue
/// \param groupIds
/// \param targetTime
/// \param groupCount
/// We must report to Master that which elements are similar to our target
void report2Master(int rank, int num_procs, deque<blockInfo> blockQueue, int targetTime, int* isSimilar){
    start_t = MPI_Wtime();

    int number_amount = 0;

    int TBLOCKNUMXYZ = targetTime * BLOCKNUMXYZ;

    deque<blockInfo>::iterator it;

    if(rank == MASTER){
        // Master is different, it needs to collect all similar elements
        // And prepare to use region growing algorithm to group them
        memset(isSimilar, -1, sizeof(int) * BLOCKNUMXYZW);

        int index;
        for(it = (blockQueue).begin(); it != (blockQueue).end(); ++it){
            if((*it).similar){
                index = ((*it).t - targetTime) * BLOCKNUMXYZ + (*it).lindex;

                isSimilar[index] = 1;
            }
        }
    }else{
        // Slaves just collect the gids of the similar ones
        // and be prepared to send to Master
        for(it = (blockQueue).begin(); it != (blockQueue).end(); ++it){
            if((*it).similar){
                isSimilar[number_amount] = getGlobalId((*it).t, (*it).lindex);
                number_amount++;
            }
        }
    }

    end_t = MPI_Wtime();

    computing_t += end_t - start_t;

    start_t = MPI_Wtime();

    int i, j, lindexInWindow;

    if(rank == MASTER){
        MPI_Status status;

        int* recv = new int[BLOCKNUMXYZW];

        for(i = 0; i < num_procs - 1; ++i){
            MPI_Recv(recv, BLOCKNUMXYZW, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);

            MPI_Get_count(&status, MPI_INT, &number_amount);

            for(j = 0; j < number_amount; ++j){
                lindexInWindow = recv[j] - TBLOCKNUMXYZ;

                isSimilar[lindexInWindow] = 1;
            }
        }

        delete[] recv;
    }else{

    //    printf("sender %d sends messages %d \n", rank, number_amount);
        MPI_Send(isSimilar, number_amount, MPI_INT, MASTER, rank, MPI_COMM_WORLD);
    }

    end_t = MPI_Wtime();

    sendrecv_t += end_t - start_t;
}

bool shouldInAGroup(int* isSimilar, int* groupIds, int offset2, int index_similar){
    int index;

    if(isSimilar[index_similar] == 1){
        index = index_similar + offset2;

        if(index >= BLOCKNUMXYZW){
            index -= BLOCKNUMXYZW;
        }

        if(groupIds[index] == -1){
            return true;
        }
    }

    return false;
}

void regionGrowing(int* isSimilar, int* groupIds, int roundStarter, int currGroupIndex, int targetGid, int targetTime, int targetIndex, int lastTimeInWindow){
    int offset1 = targetTime * BLOCKNUMXYZ;

    int offset2 = roundStarter * BLOCKNUMXYZ;

    //printf("targetTime, roundStarter offset1 offset2 %d %d %d %d \n", targetTime, roundStarter, offset1, offset2);

    deque<blockInGroup> myqueue;

    assert(isSimilar[targetIndex] != -1);

    blockInGroup target;
    target.lindex = targetIndex;
    target.t = targetTime;
    target.gid = targetGid;

    myqueue.push_back(target);

    isSimilar[target.gid - offset1] = 2;

    int index_similar;
    bool isIn;

    blockInGroup head;

    int bx, by, bz, temp;

    while(!myqueue.empty()){
        head = myqueue.front();

        bz = head.lindex / BLOCKNUMXY;
        temp = head.lindex % BLOCKNUMXY;

        by = temp / BLOCKNUMX;
        bx = temp % BLOCKNUMX;

        // x: +1 -1
        if(bx + 1 < BLOCKNUMX){
            index_similar = head.gid + 1 - offset1;

            isIn = shouldInAGroup(isSimilar, groupIds, offset2, index_similar);

            if(isIn){
                blockInGroup member;
                member.lindex = head.lindex + 1;
                member.t = head.t;
                member.gid = head.gid + 1;

                myqueue.push_back(member);

                isSimilar[index_similar] = 2;
            }
        }

        if(bx > 0){
            index_similar = head.gid - 1 - offset1;

            isIn = shouldInAGroup(isSimilar, groupIds, offset2, index_similar);

            if(isIn){
                blockInGroup member;
                member.lindex = head.lindex - 1;
                member.t = head.t;
                member.gid = head.gid - 1;

                myqueue.push_back(member);

                isSimilar[index_similar] = 2;
            }
        }

        // y: +1 -1
        if(by + 1 < BLOCKNUMY){
            index_similar = head.gid + BLOCKNUMX - offset1;

            isIn = shouldInAGroup(isSimilar, groupIds, offset2, index_similar);

            if(isIn){
                blockInGroup member;
                member.lindex = head.lindex + BLOCKNUMX;
                member.t = head.t;
                member.gid = head.gid + BLOCKNUMX;

                myqueue.push_back(member);

                isSimilar[index_similar] = 2;
            }
        }

        if(by > 0){
            index_similar = head.gid - BLOCKNUMX - offset1;

            isIn = shouldInAGroup(isSimilar, groupIds, offset2, index_similar);

            if(isIn){
                blockInGroup member;
                member.lindex = head.lindex - BLOCKNUMX;
                member.t = head.t;

                member.gid = head.gid - BLOCKNUMX;

                myqueue.push_back(member);

                isSimilar[index_similar] = 2;
            }
        }

        // z: +1 -1
        if(bz + 1 < BLOCKNUMZ){
            index_similar = head.gid + BLOCKNUMXY - offset1;

            isIn = shouldInAGroup(isSimilar, groupIds, offset2, index_similar);

            if(isIn){
                blockInGroup member;
                member.lindex = head.lindex + BLOCKNUMXY;
                member.t = head.t;
                member.gid = head.gid + BLOCKNUMXY;

                myqueue.push_back(member);

                isSimilar[index_similar] = 2;
            }
        }

        if(bz > 0){
            index_similar = head.gid - BLOCKNUMXY - offset1;

            isIn = shouldInAGroup(isSimilar, groupIds, offset2, index_similar);

            if(isIn){
                blockInGroup member;
                member.lindex = head.lindex - BLOCKNUMXY;
                member.t = head.t;
                member.gid = head.gid - BLOCKNUMXY;

                myqueue.push_back(member);

                isSimilar[index_similar] = 2;
            }
        }

        // t: +1

        if(head.t + 1 <= lastTimeInWindow){
            index_similar = head.gid + BLOCKNUMXYZ - offset1;

            isIn = shouldInAGroup(isSimilar, groupIds, offset2, index_similar);

            if(isIn){
                blockInGroup member;
                member.lindex = head.lindex;
                member.t = head.t + 1;
                member.gid = head.gid + BLOCKNUMXYZ;

                myqueue.push_back(member);

                isSimilar[index_similar] = 2;
            }
        }

        myqueue.pop_front();
    }

    // now myqueue is empty, all new members are isSimilar == 2
    for(int i = 0; i < BLOCKNUMXYZW; ++i){
        if(isSimilar[i] == 2){
            int index;

            index = i + offset2;

            if(index >= BLOCKNUMXYZW){
                index -= BLOCKNUMXYZW;
            }

            assert(groupIds[index] == -1);

            groupIds[index] = currGroupIndex;
        }
    }
}

void saveTargets(vector<int> targetRecorder){
    char fn[256];

    sprintf(fn, "%s/clusterIds.txt", RESULTPATH);

    FILE* pFile;

    pFile = fopen(fn, "wt");

    if(pFile == NULL){
        printf("can not open file %s \n",fn);

        exit(1);
    }else{
        for(int i = 0; i < targetRecorder.size(); ++i){
            fprintf(pFile, "%d\n", targetRecorder[i]);
        }

        fclose(pFile);
    }
}

int findTheNextTarget(int* groupIds, int startRound, int targetTime, int lastTimeInWindow, int targetIndex){
    int i, j, t, newTarget;

    for(i = targetTime; i <= lastTimeInWindow; ++i){
        if(i == targetTime){
            for(j = targetIndex; j < BLOCKNUMXYZ; ++j){
                if(groupIds[startRound * BLOCKNUMXYZ + j] == -1){
                    newTarget = i * BLOCKNUMXYZ + j;

                    return newTarget;
                }
            }
        }else{
            t = i - targetTime + startRound;

            if(t >= TIMEWINDOW){
                t -= TIMEWINDOW;
            }

            for(j = 0; j < BLOCKNUMXYZ; ++j){
                if(groupIds[t * BLOCKNUMXYZ + j] == -1){
                    newTarget = i * BLOCKNUMXYZ + j;

                    return newTarget;
                }
            }
        }
    }

    if(lastTimeInWindow + 1 >= BLOCKNUMT){
        return -1;
    }else{
        return ((lastTimeInWindow + 1) * BLOCKNUMXYZ);
    }
}

int saveGroupId(int* groupIds, int startRound, int targetTime, int newT){
    // save and reset to -1 for future purpose
    char fn[256];
    FILE* pFile;

    int i, curindex;

    int* myData = new int[BLOCKNUMXYZ];

    for(i = targetTime; i < newT; ++i){
        sprintf(fn, "%s/clusterIds/%04d.dat", RESULTPATH, i);

        pFile = fopen(fn, "wb");

        if(pFile == NULL){
            printf("can not open file %s\n", fn);

            exit(1);
        }else{
            curindex = i - targetTime + startRound;

            if(curindex >= TIMEWINDOW){
                curindex -= TIMEWINDOW;
            }

            for(int z = 0; z < BLOCKNUMXYZ; ++z){
                myData[z] = groupIds[curindex * BLOCKNUMXYZ + z];
            }

            //fwrite(&(groupIds[curindex * BLOCKNUMXYZ]), sizeof(int), BLOCKNUMXYZ, pFile);

            fwrite(myData, sizeof(int), BLOCKNUMXYZ, pFile);

            fclose(pFile);
        }

        sprintf(fn, "%s/clusterIds/%04d.txt", RESULTPATH, i);

        pFile = fopen(fn, "w");

        if(pFile == NULL){
            printf("can not open file %s\n", fn);

            exit(1);
        }else{
            curindex = i - targetTime + startRound;

            if(curindex >= TIMEWINDOW){
                curindex -= TIMEWINDOW;
            }

            //fwrite(&(groupIds[curindex * BLOCKNUMXYZ]), sizeof(int), BLOCKNUMXYZ, pFile);

            for(int z = 0; z < BLOCKNUMXYZ; ++z){
                //fprintf(pFile, "%d\n", groupIds[curindex * BLOCKNUMXYZ + z]);
                fprintf(pFile, "%d\n", myData[z]);
            }

            fclose(pFile);

        //    memset(&(groupIds[curindex * BLOCKNUMXYZ]), -1, sizeof(int) * BLOCKNUMXYZ);
        }

        memset(&(groupIds[curindex * BLOCKNUMXYZ]), -1, sizeof(int) * BLOCKNUMXYZ);
    }

    startRound += newT - targetTime;

    if(startRound >= TIMEWINDOW){
        startRound -= TIMEWINDOW;
    }

    delete[] myData;

    return startRound;
}

///////////////////////////////////
int main(int argc, char **argv) {
    int i;

    //////// Let us begin
    int rank, num_procs;

    MPI_Init(&argc, &argv);

    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    MPI_Comm_size(MPI_COMM_WORLD, &num_procs);

    /////////////////////////////
    loading_t = 0;
    computing_t = 0;
    saving_t = 0;
    broadcast_t = 0;
    sendrecv_t = 0;
    regionGrowing_t = 0;

    totalstart_t = MPI_Wtime();
    ////////////////////////////

    ////////////////////
    // now let's allocate blocks in each time step to processes
    int num_blocks = BLOCKNUMXYZ / num_procs;
    int num_remain = BLOCKNUMXYZ - num_procs * num_blocks;

    //assert(num_remain >= 0);

    if(num_remain != 0){
        if(rank < num_remain){
            num_blocks++;
        }
    }

    int* blockIDs = new int[num_blocks];

    //assert(num_blocks > 0);

    blockIDs[0] = rank;
    for(i = 1; i < num_blocks; ++i){
        blockIDs[i] = blockIDs[i - 1] + num_procs;
    }

    /////////////////////
    int targetGID = 0;
    int targetIndex;
    float targetHis[BINSIZE];
    int targetTime;
    int lastTimeInWindow;

    int groupCount = 0;

    int* groupIds = NULL;

    if(rank == MASTER){
        groupIds = new int[BLOCKNUMXYZW];
        memset(groupIds, -1, sizeof(int) * BLOCKNUMXYZW);
    }

    int startRound = 0;

    deque<blockInfo> blockQueue;

    vector<int> targetRecorder;

    targetRecorder.push_back(targetGID);

    ///////////////////////
    start_t = MPI_Wtime();

    MPI_Bcast(&targetGID, 1, MPI_INT, MASTER, MPI_COMM_WORLD);

    end_t = MPI_Wtime();

    broadcast_t += end_t - start_t;
    /////////////////////////

    bool containTarget;

    int* isSimilar = new int[BLOCKNUMXYZW];

    int lastLoadedTime = -1;

    while(targetGID != -1){
        start_t = MPI_Wtime();

        getTargetTimeAndLastTimeInWindow(targetGID, &targetTime, &targetIndex, &lastTimeInWindow);

        removeOldBlocks(&blockQueue, targetGID);

        end_t = MPI_Wtime();

        computing_t += end_t - start_t;

        addNewBlocks(&blockQueue, lastTimeInWindow, blockIDs, num_blocks, &lastLoadedTime);

        //printf("after add new blocks -- rank %d \n", rank);

        start_t = MPI_Wtime();

        //bool haveTarget(deque<blockInfo> blockQueue, int targetTime, int targetIndex, float* targetHis){
        containTarget = haveTarget(blockQueue, targetTime, targetIndex, targetHis);

        end_t = MPI_Wtime();

        computing_t += end_t - start_t;

        //printf("Who contains target %d -- rank %d \n", containTarget, rank);

        start_t = MPI_Wtime();

        broadCastHis(rank, num_procs, containTarget, targetHis);

        end_t = MPI_Wtime();

        sendrecv_t += end_t - start_t;

        //printf("After broadcast histogram -- rank %d \n", rank);

        start_t = MPI_Wtime();

        compare(targetHis, &blockQueue);

        end_t = MPI_Wtime();

        computing_t += end_t - start_t;

        //printf("After comparison -- rank %d \n", rank);

        MPI_Barrier(MPI_COMM_WORLD);

        //void report2Master(int rank, int num_procs, deque<blockInfo> blockQueue, int targetTime, int* isSimilar){
        report2Master(rank, num_procs, blockQueue, targetTime, isSimilar);

        //printf("After report to Master -- rank %d \n", rank);

        if(rank == MASTER){
            start_t = MPI_Wtime();

            //void regionGrowing(int* isSimilar, int* groupIds, int roundStarter, int currGroupIndex, int targetGid, int targetTime, int targetIndex, int lastTimeInWindow)
            regionGrowing(isSimilar, groupIds, startRound, groupCount, targetGID, targetTime, targetIndex, lastTimeInWindow);

            //printf("Master region growing\n");

            end_t = MPI_Wtime();

            regionGrowing_t += end_t - start_t;

            start_t = MPI_Wtime();

            //int findTheNextTarget(int* groupIds, int startRound, int targetTime, int lastTimeInWindow, int targetIndex){
            targetGID = findTheNextTarget(groupIds, startRound, targetTime, lastTimeInWindow, targetIndex);

            targetRecorder.push_back(targetGID);

            //printf("Master find next target %d\n", targetGID);

            int newT;

            if(targetGID == -1){
                newT = BLOCKNUMT;
            }else{
                newT = targetGID / BLOCKNUMXYZ;
            }

            end_t = MPI_Wtime();

            regionGrowing_t += end_t - start_t;

            start_t = MPI_Wtime();

            startRound = saveGroupId(groupIds, startRound, targetTime, newT);

            end_t = MPI_Wtime();

            saving_t += end_t - start_t;

            //printf("Master save group ids \n");
        }

        MPI_Barrier(MPI_COMM_WORLD);

        start_t = MPI_Wtime();

        // begin a new round
        MPI_Bcast(&targetGID, 1, MPI_INT, MASTER, MPI_COMM_WORLD);

        end_t = MPI_Wtime();

        broadcast_t += end_t - start_t;

        if(rank == 0)
            printf("After broad cast new target %d -- total %d \n", targetGID, targetRecorder.size());

        groupCount++;
    }

    delete[] isSimilar;

    if(rank == MASTER){
        start_t = MPI_Wtime();

        saveTargets(targetRecorder);

        end_t = MPI_Wtime();

        saving_t += end_t - start_t;
    }

    ////////////////////////////
    totalend_t = MPI_Wtime();

    totalTime_t = totalend_t - totalstart_t;

    printf("%d th process, totalTime loading computing saving growing broadcast sendrecv %f; %f; %f; %f; %f; %f; %f; \n",
           rank, totalTime_t, loading_t, computing_t, saving_t, regionGrowing_t, broadcast_t, sendrecv_t);

    ////////////////////////////
    fflush(stderr);

    MPI_Barrier(MPI_COMM_WORLD);

    if (rank == MASTER)
        fprintf(stderr, "\n---Completed successfully---\n");

    MPI_Finalize();

    return 0;
}
